@@ -1,6 +1,6 @@
-/*	$Id: diff.c,v 1.16 2017/10/18 00:04:10 kristaps Exp $ */
+
 /*
- * Copyright (c) 2017 Kristaps Dzonsons <kristaps@bsd.lv>
+ * Copyright (c) 2017, 2018 Kristaps Dzonsons <kristaps@bsd.lv>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -16,25 +16,29 @@
  */
 #include "config.h"
 
-#include <sys/queue.h>
+#if HAVE_SYS_QUEUE
+# include <sys/queue.h>
+#endif
 #include <sys/types.h>
 
 #include <assert.h>
-#if HAVE_ERR
-# include <err.h>
-#endif
+#include <ctype.h>
 #include <float.h>
 #include <math.h>
 #if HAVE_MD5
 # include <md5.h>
 #endif
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "lowdown.h"
+#include "libdiff.h"
 #include "extern.h"
 
+#define	DEBUG 0
+
 struct	xnode {
 	char		 sig[MD5_DIGEST_STRING_LENGTH]; /* signature */
 	double		 weight; /* priority queue weight */
@@ -58,11 +62,24 @@ struct	pnode {
 
 TAILQ_HEAD(pnodeq, pnode);
 
-#if 0
+/*
+ * A node used in computing the shortest edit script.
+ */
+struct	sesnode {
+	char		*buf; /* buffer */
+	size_t		 bufsz; /* length of buffer (less NUL) */
+	int		 tailsp; /* whether there's trailing space */
+	int		 headsp; /* whether there's leading space */
+};
+
+#if DEBUG
 static	const char *const names[LOWDOWN__MAX] = {
 	"LOWDOWN_ROOT",			/* LOWDOWN_ROOT */
 	"LOWDOWN_BLOCKCODE",            /* LOWDOWN_BLOCKCODE */
 	"LOWDOWN_BLOCKQUOTE",           /* LOWDOWN_BLOCKQUOTE */
+	"LOWDOWN_DEFINITION",		/* LOWDOWN_DEFINITION */
+	"LOWDOWN_DEFINITION_TITLE",	/* LOWDOWN_DEFINITION_TITLE */
+	"LOWDOWN_DEFINITION_DATA",	/* LOWDOWN_DEFINITION_DATA */
 	"LOWDOWN_HEADER",               /* LOWDOWN_HEADER */
 	"LOWDOWN_HRULE",                /* LOWDOWN_HRULE */
 	"LOWDOWN_LIST",                 /* LOWDOWN_LIST */
@@ -93,16 +110,16 @@ static	const char *const names[LOWDOWN__MAX] = {
 	"LOWDOWN_ENTITY",               /* LOWDOWN_ENTITY */
 	"LOWDOWN_NORMAL_TEXT",          /* LOWDOWN_NORMAL_TEXT */
 	"LOWDOWN_DOC_HEADER",           /* LOWDOWN_DOC_HEADER */
+	"LOWDOWN_META",           	/* LOWDOWN_META */
 	"LOWDOWN_DOC_FOOTER",           /* LOWDOWN_DOC_FOOTER */
 };
 #endif
 
-
 static void
-MD5Updatebuf(MD5_CTX *ctx, const hbuf *v)
+MD5Updatebuf(MD5_CTX *ctx, const struct lowdown_buf *v)
 {
 
-	MD5Update(ctx, (const u_int8_t *)v->data, v->size);
+	MD5Update(ctx, (const uint8_t *)v->data, v->size);
 }
 
 static void
@@ -134,15 +151,15 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 	/* Get our node slot. */
 
 	if (n->id >= map->maxsize) {
-		map->nodes = xreallocarray
-			(map->nodes, n->id + 64,
+		map->nodes = xrecallocarray
+			(map->nodes, 
+			 map->maxsize, n->id + 64,
 			 sizeof(struct xnode));
 		map->maxsize = n->id + 64;
 	}
 
 	assert(n->id < map->maxsize);
 	xn = &map->nodes[n->id];
-	memset(xn, 0, sizeof(struct xnode));
 	assert(NULL == xn->node);
 	assert(0.0 == xn->weight);
 	xn->node = n;
@@ -285,7 +302,7 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 		MD5Updatebuf(&ctx, &n->rndr_image.alt);
 		break;
 	case LOWDOWN_MATH_BLOCK:
-		MD5Updatev(&ctx, &n->rndr_math.displaymode, 
+		MD5Updatev(&ctx, &n->rndr_math.blockmode, 
 			sizeof(int));
 		break;
 	case LOWDOWN_BLOCKHTML:
@@ -298,7 +315,7 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 	MD5End(&ctx, xn->sig);
 
 	if (NULL != parent)
-		MD5Update(parent, (u_int8_t *)xn->sig, 
+		MD5Update(parent, (uint8_t *)xn->sig, 
 			MD5_DIGEST_STRING_LENGTH - 1);
 
 	if (xn->weight > map->maxweight)
@@ -405,7 +422,7 @@ candidate(struct xnode *xnew, struct xmap *xnewmap,
 	assert(NULL != xnew->node);
 	assert(NULL != xold->node);
 
-	if (NULL == xnew->match) {
+	if (NULL == xnew->optmatch) {
 		xnew->optmatch = xold->node;
 		xnew->opt = optimality
 			(xnew, xnewmap, xold, xoldmap);
@@ -420,7 +437,7 @@ candidate(struct xnode *xnew, struct xmap *xnewmap,
 		 * Choose the lesser of the norms.
 		 */
 		dold = llabs((long long)
-			(xnew->match->id - xnew->node->id));
+			(xnew->optmatch->id - xnew->node->id));
 		dnew = llabs((long long)
 			(xold->node->id - xnew->node->id));
 		if (dold > dnew) {
@@ -430,7 +447,7 @@ candidate(struct xnode *xnew, struct xmap *xnewmap,
 	} else if (opt > xnew->opt) {
 		xnew->optmatch = xold->node;
 		xnew->opt = opt;
-	}
+	} 
 }
 
 /*
@@ -579,8 +596,7 @@ match_down(struct xnode *xnew, struct xmap *xnewmap,
 static struct lowdown_node *
 node_clone(const struct lowdown_node *v, size_t id)
 {
-	struct lowdown_node *n;
-	size_t		 i;
+	struct lowdown_node	*n;
 
 	n = xcalloc(1, sizeof(struct lowdown_node));
 	TAILQ_INIT(&n->children);
@@ -588,20 +604,9 @@ node_clone(const struct lowdown_node *v, size_t id)
 	n->id = id;
 
 	switch (n->type) {
-	case LOWDOWN_DOC_HEADER:
-		n->rndr_doc_header.msz =
-			v->rndr_doc_header.msz;
-		if (0 == n->rndr_doc_header.msz)
-			break;
-		n->rndr_doc_header.m = xcalloc
-			(v->rndr_doc_header.msz,
-			 sizeof(struct lowdown_meta));
-		for (i = 0; i < n->rndr_doc_header.msz; i++) {
-			n->rndr_doc_header.m[i].key = xstrdup
-				(v->rndr_doc_header.m[i].key);
-			n->rndr_doc_header.m[i].value = xstrdup
-				(v->rndr_doc_header.m[i].value);
-		}
+	case LOWDOWN_META:
+		hbuf_clone(&v->rndr_meta.key,
+			&n->rndr_meta.key);
 		break;
 	case LOWDOWN_LIST:
 		n->rndr_list.flags = v->rndr_list.flags;
@@ -671,8 +676,8 @@ node_clone(const struct lowdown_node *v, size_t id)
 			&n->rndr_image.alt);
 		break;
 	case LOWDOWN_MATH_BLOCK:
-		n->rndr_math.displaymode = 
-			v->rndr_math.displaymode;
+		n->rndr_math.blockmode = 
+			v->rndr_math.blockmode;
 		break;
 	case LOWDOWN_BLOCKHTML:
 		hbuf_clone(&v->rndr_blockhtml.text,
@@ -707,6 +712,179 @@ node_clonetree(const struct lowdown_node *v, size_t *id)
 	return(n);
 }
 
+/*
+ * Count the number of words in a normal-text node.
+ */
+static size_t
+node_countwords(const struct lowdown_node *n)
+{
+	const char	*cp;
+	size_t		 i = 0, sz, words = 0;
+
+	cp = n->rndr_normal_text.text.data;
+	sz = n->rndr_normal_text.text.size;
+
+	/* Skip leading space. */
+
+	while (i < sz &&
+	       isspace((unsigned char)cp[i]))
+		i++;
+
+	/* First go through word, then trailing space. */
+
+	while (i < sz) {
+		assert( ! isspace((unsigned char)cp[i]));
+		words++;
+		while (i < sz &&
+		       ! isspace((unsigned char)cp[i]))
+			i++;
+		while (i < sz && 
+		       isspace((unsigned char)cp[i]))
+			i++;
+	}
+
+	return words;
+}
+
+/*
+ * Like node_countwords(), except dupping individual words into a
+ * structure.
+ */
+static void
+node_tokenise(const struct lowdown_node *n, 
+	struct sesnode *toks, size_t toksz, char **savep)
+{
+	char	*cp;
+	size_t	 i = 0, sz, words = 0;
+
+	*savep = NULL;
+
+	if (0 == toksz)
+		return;
+
+	sz = n->rndr_normal_text.text.size;
+	*savep = cp = xmalloc(sz + 1);
+	memcpy(cp, n->rndr_normal_text.text.data, sz);
+	cp[sz] = '\0';
+
+	*savep = cp;
+
+	/* Skip leading space. */
+
+	if (i < sz)
+		toks[0].headsp = isspace((unsigned char)cp[0]);
+
+	while (i < sz &&
+	       isspace((unsigned char)cp[i]))
+		i++;
+
+	while (i < sz) {
+		assert(words < toksz);
+		assert( ! isspace((unsigned char)cp[i]));
+		toks[words].buf = &cp[i];
+		toks[words].bufsz = 0;
+		while (i < sz &&
+		       ! isspace((unsigned char)cp[i])) {
+			toks[words].bufsz++;
+			i++;
+		}
+		words++;
+		if (i == sz)
+			break;
+		toks[words - 1].tailsp = 1;
+		assert(isspace((unsigned char)cp[i]));
+		cp[i++] = '\0';
+		while (i < sz && 
+		       isspace((unsigned char)cp[i]))
+			i++;
+	}
+}
+
+static int
+node_word_cmp(const void *p1, const void *p2)
+{
+	const struct sesnode *l1 = p1, *l2 = p2;
+
+	if (l1->bufsz != l2->bufsz)
+		return 0;
+	return 0 == strncmp(l1->buf, l2->buf, l1->bufsz);
+}
+
+static void
+node_lcs(const struct lowdown_node *nold,
+	const struct lowdown_node *nnew,
+	struct lowdown_node *n, size_t *id)
+{
+	const struct sesnode *tmp;
+	struct lowdown_node *nn;
+	struct sesnode	*newtok, *oldtok;
+	char		*newtokbuf, *oldtokbuf;
+	size_t		 i, newtoksz, oldtoksz;
+	struct diff	 d;
+
+	newtoksz = node_countwords(nnew);
+	oldtoksz = node_countwords(nold);
+
+	newtok = xcalloc(newtoksz, sizeof(struct sesnode));
+	oldtok = xcalloc(oldtoksz, sizeof(struct sesnode));
+
+	node_tokenise(nnew, newtok, newtoksz, &newtokbuf);
+	node_tokenise(nold, oldtok, oldtoksz, &oldtokbuf);
+
+	diff(&d, node_word_cmp, sizeof(struct sesnode), 
+		oldtok, oldtoksz, newtok, newtoksz);
+
+	for (i = 0; i < d.sessz; i++) {
+		tmp = d.ses[i].e;
+
+		if (tmp->headsp) {
+			nn = xcalloc(1, sizeof(struct lowdown_node));
+			TAILQ_INIT(&nn->children);
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nn->type = LOWDOWN_NORMAL_TEXT;
+			nn->id = (*id)++;
+			nn->parent = n;
+			nn->rndr_normal_text.text.size = 1;
+			nn->rndr_normal_text.text.data = xstrdup(" ");
+		}
+
+		nn = xcalloc(1, sizeof(struct lowdown_node));
+		TAILQ_INIT(&nn->children);
+		TAILQ_INSERT_TAIL(&n->children, nn, entries);
+		nn->type = LOWDOWN_NORMAL_TEXT;
+		nn->id = (*id)++;
+		nn->parent = n;
+		nn->rndr_normal_text.text.size = tmp->bufsz;
+		nn->rndr_normal_text.text.data = 
+			xcalloc(1, tmp->bufsz + 1);
+		memcpy(nn->rndr_normal_text.text.data,
+			tmp->buf, tmp->bufsz);
+		nn->chng = DIFF_DELETE == d.ses[i].type ?
+			LOWDOWN_CHNG_DELETE :
+			DIFF_ADD == d.ses[i].type ?
+			LOWDOWN_CHNG_INSERT :
+			LOWDOWN_CHNG_NONE;
+
+		if (tmp->tailsp) {
+			nn = xcalloc(1, sizeof(struct lowdown_node));
+			TAILQ_INIT(&nn->children);
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nn->type = LOWDOWN_NORMAL_TEXT;
+			nn->id = (*id)++;
+			nn->parent = n;
+			nn->rndr_normal_text.text.size = 1;
+			nn->rndr_normal_text.text.data = xstrdup(" ");
+		}
+	}
+
+	free(d.ses);
+	free(d.lcs);
+	free(newtok);
+	free(oldtok);
+	free(newtokbuf);
+	free(oldtokbuf);
+}
+
 /*
  * Merge the new tree "nnew" with the old "nold" using a depth-first
  * algorithm.
@@ -740,7 +918,7 @@ node_merge(const struct lowdown_node *nold,
 
 	assert(xnew->match == xold->node);
 
-	n = node_clone(nnew, *id++);
+	n = node_clone(nnew, (*id)++);
 
 	/* Now walk through the children on both sides. */
 
@@ -758,7 +936,8 @@ node_merge(const struct lowdown_node *nold,
 
 		while (NULL != nold) {
 			xold = &xoldmap->nodes[nold->id];
-			if (NULL != xold->match) 
+			if (NULL != xold->match ||
+			    LOWDOWN_NORMAL_TEXT == nold->type)
 				break;
 			nn = node_clonetree(nold, id);
 			nn->parent = n;
@@ -774,6 +953,45 @@ node_merge(const struct lowdown_node *nold,
 		 * (not necessarily at this level) with the old.
 		 */
 
+		while (NULL != nnew) {
+			xnew = &xnewmap->nodes[nnew->id];
+			if (NULL != xnew->match ||
+			    LOWDOWN_NORMAL_TEXT == nnew->type)
+				break;
+			nn = node_clonetree(nnew, id);
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nn->parent = n;
+			nn->chng = LOWDOWN_CHNG_INSERT;
+			nnew = TAILQ_NEXT(nnew, entries);
+		}
+
+		/*
+		 * If both nodes are text nodes, then we want to run the
+		 * LCS algorithm on them.
+		 * This is an extension of the BULD algorithm.
+		 */
+
+		if (NULL != nold && NULL != nnew &&
+		    LOWDOWN_NORMAL_TEXT == nold->type &&
+		    NULL == xold->match &&
+		    LOWDOWN_NORMAL_TEXT == nnew->type &&
+		    NULL == xnew->match) {
+			node_lcs(nold, nnew, n, id);
+			nold = TAILQ_NEXT(nold, entries);
+			nnew = TAILQ_NEXT(nnew, entries);
+		}
+
+		while (NULL != nold) {
+			xold = &xoldmap->nodes[nold->id];
+			if (NULL != xold->match)
+				break;
+			nn = node_clonetree(nold, id);
+			nn->parent = n;
+			nn->chng = LOWDOWN_CHNG_DELETE;
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			nold = TAILQ_NEXT(nold, entries);
+		}
+
 		while (NULL != nnew) {
 			xnew = &xnewmap->nodes[nnew->id];
 			if (NULL != xnew->match)
@@ -865,10 +1083,10 @@ node_merge(const struct lowdown_node *nold,
 		nold = TAILQ_NEXT(nold, entries);
 	}
 
-	return(n);
+	return n;
 }
 
-#if 0
+#if DEBUG
 static void
 node_print(const struct lowdown_node *n, const struct xmap *map, size_t ind)
 {
@@ -894,20 +1112,97 @@ node_print(const struct lowdown_node *n, const struct xmap *map, size_t ind)
 #endif
 
 /*
- * Our optimising phase is a bit complicated.
- * It is not reflect in any part of the paper.
- * In short, we want to look, from bottom-up, at all nodes that aren't
- * matched.
- * We then examine all the children of the un-matched nodes and see
- * which of their matches, if found, are under a root that's the same
- * node as we are.
- * This lets us compuate the largest fraction of un-matched nodes'
+ * Optimise from top down.
+ * This works by selecting matching non-terminal nodes, both adjacent
+ * (i.e., children of the same adjacent nodes), and seeing if their
+ * immediate siblings may be matched by label.
+ * This works well when looking at pure-paragraph changes.
+ */
+static void
+node_optimise_topdown(const struct lowdown_node *n, 
+	struct xmap *newmap, struct xmap *oldmap)
+{
+	struct xnode	*xn, *xmatch, *xnchild, *xmchild,
+			*xnnext, *xmnext;
+	const struct lowdown_node 
+			*match, *nchild, *mchild, *nnext, 
+			*mnext;
+
+	if (TAILQ_EMPTY(&n->children))
+		return;
+
+	xn = &newmap->nodes[n->id];
+	assert(NULL != xn);
+
+	if (NULL == (match = xn->match))
+		return;
+
+	xmatch = &oldmap->nodes[match->id];
+	assert(NULL != xmatch);
+
+	TAILQ_FOREACH(nchild, &n->children, entries) {
+		if (TAILQ_EMPTY(&nchild->children))
+			continue;
+		xnchild = &newmap->nodes[nchild->id];
+		assert(NULL != xnchild);
+		if (NULL == (mchild = xnchild->match))
+			continue;
+		if (mchild->parent->id != match->id)
+			continue;
+		xmchild = &oldmap->nodes[mchild->id];
+		assert(NULL != xmchild);
+
+		/* 
+		 * Do we have a non-terminal sibling after us without a
+		 * match? 
+		 */
+
+		if (NULL == (nnext = TAILQ_NEXT(nchild, entries)))
+			continue;
+		if (TAILQ_EMPTY(&nnext->children))
+			continue;
+		xnnext = &newmap->nodes[nnext->id];
+		assert(NULL != xnnext);
+		if (NULL != xnnext->match)
+			continue;
+
+		if (NULL == (mnext = TAILQ_NEXT(mchild, entries)))
+			continue;
+		if (TAILQ_EMPTY(&mnext->children))
+			continue;
+		xmnext = &oldmap->nodes[mnext->id];
+		assert(NULL != xmnext);
+		if (NULL != xmnext->match)
+			continue;
+
+		if ( ! match_eq(nnext, mnext))
+			continue;
+
+#if DEBUG
+		warnx("%s (%zu): subtree match: %s (%zu) -> %zu",
+			names[n->type], n->id, 
+			names[nchild->type], nchild->id, mchild->id);
+#endif
+
+		xnnext->match = mnext;
+		xmnext->match = nnext;
+	}
+
+	TAILQ_FOREACH(nchild, &n->children, entries)
+		node_optimise_topdown(nchild, newmap, oldmap);
+}
+
+/*
+ * Optimise bottom-up over all un-matched nodes: examine all the
+ * children of the un-matched nodes and see which of their matches, if
+ * found, are under a root that's the same node as we are.
+ * This lets us compute the largest fraction of un-matched nodes'
  * children that are in the same tree.
  * If that fraction is >50%, then we consider that the subtrees are
  * matched.
  */
 static void
-node_optimise(const struct lowdown_node *n, 
+node_optimise_bottomup(const struct lowdown_node *n, 
 	struct xmap *newmap, struct xmap *oldmap)
 {
 	const struct lowdown_node *nn, *on, *nnn, *maxn = NULL;
@@ -920,7 +1215,7 @@ node_optimise(const struct lowdown_node *n,
 
 	TAILQ_FOREACH(nn, &n->children, entries) {
 		tw += newmap->nodes[nn->id].weight;
-		node_optimise(nn, newmap, oldmap);
+		node_optimise_bottomup(nn, newmap, oldmap);
 	}
 
 	/*
@@ -984,14 +1279,9 @@ node_optimise(const struct lowdown_node *n,
 	}
 }
 
-/*
- * Algorithm: Detecting Changes in XML Documents.
- * Gregory Cobena, Serge Abiteboul, Amelie Marian.
- * https://www.cs.rutgers.edu/~amelie/papers/2002/diff.pdf
- */
 struct lowdown_node *
 lowdown_diff(const struct lowdown_node *nold,
-	const struct lowdown_node *nnew)
+	const struct lowdown_node *nnew, size_t *maxn)
 {
 	struct xmap	 xoldmap, xnewmap;
 	struct xnode	*xnew, *xold;
@@ -1044,10 +1334,13 @@ lowdown_diff(const struct lowdown_node *nold,
 
 		for (i = 0; i < xoldmap.maxid + 1; i++) {
 			xold = &xoldmap.nodes[i];
+			if (NULL == xold->node)
+				continue;
 			if (NULL != xold->match)
 				continue;
 			if (strcmp(xnew->sig, xold->sig))
 				continue;
+
 			assert(NULL == xold->match);
 			candidate(xnew, &xnewmap, xold, &xoldmap);
 		}
@@ -1082,12 +1375,12 @@ lowdown_diff(const struct lowdown_node *nold,
 	 * Our optimisation is nothing like the paper's.
 	 */
 
-#if 0
-	node_print(nnew, &xnewmap, 0);
-#endif
-	node_optimise(nnew, &xnewmap, &xoldmap);
-#if 0
+	node_optimise_topdown(nnew, &xnewmap, &xoldmap);
+	node_optimise_bottomup(nnew, &xnewmap, &xoldmap);
+
+#if DEBUG
 	node_print(nnew, &xnewmap, 0);
+	node_print(nold, &xoldmap, 0);
 #endif
 
 	/*
@@ -1099,6 +1392,10 @@ lowdown_diff(const struct lowdown_node *nold,
 	i = 0;
 	comp = node_merge(nold, &xoldmap, nnew, &xnewmap, &i);
 
+	*maxn = xnewmap.maxid > xoldmap.maxid ?
+		xnewmap.maxid + 1 :
+		xoldmap.maxid + 1;
+
 	/* Clean up and exit. */
 
 	assert(TAILQ_EMPTY(&pq));
