@@ -1,4 +1,4 @@
-/*	$Id: log.c,v 1.5 2017/07/27 15:19:19 kristaps Exp $ */
+/*	$Id: log.c,v 1.6 2019/04/01 09:39:31 kristaps Exp $ */
 /*
  * Copyright (c) 2016, Kristaps Dzonsons
  *
@@ -16,7 +16,9 @@
  */
 #include "config.h"
 
-#include <sys/queue.h>
+#if HAVE_SYS_QUEUE
+# include <sys/queue.h>
+#endif
 
 #include <stdarg.h>
 #include <stdlib.h>
